@@ -62,7 +62,7 @@ import six
 
 def cmemcache_hash(key):
     return (
-        (((binascii.crc32(key.encode('ascii')) & 0xffffffff)
+        (((binascii.crc32(key) & 0xffffffff)
           >> 16) & 0x7fff) or 1)
 serverHashFunction = cmemcache_hash
 
@@ -72,45 +72,34 @@ def useOldServerHashFunction():
     global serverHashFunction
     serverHashFunction = binascii.crc32
 
-try:
-    from zlib import compress, decompress
-    _supports_compress = True
-except ImportError:
-    _supports_compress = False
-    # quickly define a decompress just in case we recv compressed data.
-
-    def decompress(val):
-        raise _Error(
-            "Received compressed data but I don't support "
-            "compression (import error)")
-
 from io import BytesIO
-try:
-    unicode
-except NameError:
-    _has_unicode = False
+if six.PY2:
+    try:
+        unicode
+    except NameError:
+        _has_unicode = False
+    else:
+        _has_unicode = True
 else:
     _has_unicode = True
 
-try:
-    _str_cls = basestring
-except NameError:
-    _str_cls = str
+_str_cls = six.string_types
 
-valid_key_chars_re = re.compile('[\x21-\x7e\x80-\xff]+$')
+valid_key_chars_re = re.compile(b'[\x21-\x7e\x80-\xff]+$')
 
 
 #  Original author: Evan Martin of Danga Interactive
 __author__ = "Sean Reifschneider <jafo-memcached@tummy.com>"
-__version__ = "1.53"
+__version__ = "1.57"
 __copyright__ = "Copyright (C) 2003 Danga Interactive"
 #  http://en.wikipedia.org/wiki/Python_Software_Foundation_License
 __license__ = "Python Software Foundation License"
 
 SERVER_MAX_KEY_LENGTH = 250
-# Storing values larger than 1MB requires recompiling memcached.  If
-# you do, this value can be changed by doing
-# "memcache.SERVER_MAX_VALUE_LENGTH = N" after importing this module.
+# Storing values larger than 1MB requires starting memcached with -I <size> for
+# memcached >= 1.4.2 or recompiling for < 1.4.2. If you do, this value can be
+# changed by doing "memcache.SERVER_MAX_VALUE_LENGTH = N" after importing this
+# module.
 SERVER_MAX_VALUE_LENGTH = 1024 * 1024
 
 
@@ -178,6 +167,7 @@ class Client(threading.local):
 
     def __init__(self, servers, debug=0, pickleProtocol=0,
                  pickler=pickle.Pickler, unpickler=pickle.Unpickler,
+                 compressor=zlib.compress, decompressor=zlib.decompress,
                  pload=None, pid=None,
                  server_max_key_length=None, server_max_value_length=None,
                  dead_retry=_DEAD_RETRY, socket_timeout=_SOCKET_TIMEOUT,
@@ -240,6 +230,8 @@ class Client(threading.local):
         self.pickleProtocol = pickleProtocol
         self.pickler = pickler
         self.unpickler = unpickler
+        self.compressor = compressor
+        self.decompressor = decompressor
         self.persistent_load = pload
         self.persistent_id = pid
         self.server_max_key_length = server_max_key_length
@@ -257,6 +249,32 @@ class Client(threading.local):
         except TypeError:
             self.picklerIsKeyword = False
 
+    def _encode_key(self, key):
+        if isinstance(key, tuple):
+            if isinstance(key[1], six.text_type):
+                return (key[0], key[1].encode('utf8'))
+        elif isinstance(key, six.text_type):
+            return key.encode('utf8')
+        return key
+
+    def _encode_cmd(self, cmd, key, headers, noreply, *args):
+        cmd_bytes = cmd.encode() if six.PY3 else cmd
+        fullcmd = [cmd_bytes, b' ', key]
+
+        if headers:
+            if six.PY3:
+                headers = headers.encode()
+            fullcmd.append(b' ')
+            fullcmd.append(headers)
+
+        if noreply:
+            fullcmd.append(b' noreply')
+
+        if args:
+            fullcmd.append(b' ')
+            fullcmd.extend(args)
+        return b''.join(fullcmd)
+
     def reset_cas(self):
         """Reset the cas cache.
 
@@ -399,7 +417,7 @@ class Client(threading.local):
         for s in self.servers:
             s.close_socket()
 
-    def delete_multi(self, keys, time=0, key_prefix=''):
+    def delete_multi(self, keys, time=0, key_prefix='', noreply=False):
         """Delete multiple keys in the memcache doing just one query.
 
         >>> notset_keys = mc.set_multi({'a1' : 'val1', 'a2' : 'val2'})
@@ -420,6 +438,8 @@ class Client(threading.local):
         @param key_prefix: Optional string to prepend to each key when
             sending to memcache.  See docs for L{get_multi} and
             L{set_multi}.
+        @param noreply: optional parameter instructs the server to not send the
+            reply.
         @return: 1 if no failure in communication with any memcacheds.
         @rtype: int
         """
@@ -436,12 +456,13 @@ class Client(threading.local):
         for server in six.iterkeys(server_keys):
             bigcmd = []
             write = bigcmd.append
+            extra = ' noreply' if noreply else ''
             if time is not None:
                 for key in server_keys[server]:  # These are mangled keys
-                    write("delete %s %d\r\n" % (key, time))
+                    write("delete %s %d%s\r\n" % (key, time, extra))
             else:
                 for key in server_keys[server]:  # These are mangled keys
-                    write("delete %s\r\n" % key)
+                    write("delete %s%s\r\n" % (key, extra))
             try:
                 server.send_cmds(''.join(bigcmd))
             except socket.error as msg:
@@ -451,6 +472,10 @@ class Client(threading.local):
                 server.mark_dead(msg)
                 dead_servers.append(server)
 
+        # if noreply, just return
+        if noreply:
+            return rc
+
         # if any servers died on the way, don't expect them to respond.
         for server in dead_servers:
             del server_keys[server]
@@ -466,17 +491,20 @@ class Client(threading.local):
                 rc = 0
         return rc
 
-    def delete(self, key, time=0):
+    def delete(self, key, time=0, noreply=False):
         '''Deletes a key from the memcache.
 
         @return: Nonzero on success.
         @param time: number of seconds any subsequent set / update commands
         should fail. Defaults to None for no delay.
+        @param noreply: optional parameter instructs the server to not send the
+            reply.
         @rtype: int
         '''
-        return self._deletetouch(['DELETED', 'NOT_FOUND'], "delete", key, time)
+        return self._deletetouch([b'DELETED', b'NOT_FOUND'], "delete", key,
+                                 time, noreply)
 
-    def touch(self, key, time=0):
+    def touch(self, key, time=0, noreply=False):
         '''Updates the expiration time of a key in memcache.
 
         @return: Nonzero on success.
@@ -485,11 +513,14 @@ class Client(threading.local):
             unix time-since-the-epoch value. See the memcached protocol
             docs section "Storage Commands" for more info on <exptime>. We
             default to 0 == cache forever.
+        @param noreply: optional parameter instructs the server to not send the
+            reply.
         @rtype: int
         '''
-        return self._deletetouch(['TOUCHED'], "touch", key, time)
+        return self._deletetouch([b'TOUCHED'], "touch", key, time, noreply)
 
-    def _deletetouch(self, expected, cmd, key, time=0):
+    def _deletetouch(self, expected, cmd, key, time=0, noreply=False):
+        key = self._encode_key(key)
         if self.do_check_key:
             self.check_key(key)
         server, key = self._get_server(key)
@@ -497,12 +528,14 @@ class Client(threading.local):
             return 0
         self._statlog(cmd)
         if time is not None and time != 0:
-            cmd = "%s %s %d" % (cmd, key, time)
+            fullcmd = self._encode_cmd(cmd, key, str(time), noreply)
         else:
-            cmd = "%s %s" % (cmd, key)
+            fullcmd = self._encode_cmd(cmd, key, None, noreply)
 
         try:
-            server.send_cmd(cmd)
+            server.send_cmd(fullcmd)
+            if noreply:
+                return 1
             line = server.readline()
             if line and line.strip() in expected:
                 return 1
@@ -514,7 +547,7 @@ class Client(threading.local):
             server.mark_dead(msg)
         return 0
 
-    def incr(self, key, delta=1):
+    def incr(self, key, delta=1, noreply=False):
         """Increment value for C{key} by C{delta}
 
         Sends a command to the server to atomically increment the
@@ -539,12 +572,15 @@ class Client(threading.local):
         @param delta: Integer amount to increment by (should be zero
         or greater).
 
-        @return: New value after incrementing.
+        @param noreply: optional parameter instructs the server to not send the
+        reply.
+
+        @return: New value after incrementing, no None for noreply or error.
         @rtype: int
         """
-        return self._incrdecr("incr", key, delta)
+        return self._incrdecr("incr", key, delta, noreply)
 
-    def decr(self, key, delta=1):
+    def decr(self, key, delta=1, noreply=False):
         """Decrement value for C{key} by C{delta}
 
         Like L{incr}, but decrements.  Unlike L{incr}, underflow is
@@ -554,23 +590,29 @@ class Client(threading.local):
         @param delta: Integer amount to decrement by (should be zero
         or greater).
 
-        @return: New value after decrementing or None on error.
+        @param noreply: optional parameter instructs the server to not send the
+        reply.
+
+        @return: New value after decrementing,  or None for noreply or error.
         @rtype: int
         """
-        return self._incrdecr("decr", key, delta)
+        return self._incrdecr("decr", key, delta, noreply)
 
-    def _incrdecr(self, cmd, key, delta):
+    def _incrdecr(self, cmd, key, delta, noreply=False):
+        key = self._encode_key(key)
         if self.do_check_key:
             self.check_key(key)
         server, key = self._get_server(key)
         if not server:
             return None
         self._statlog(cmd)
-        cmd = "%s %s %d" % (cmd, key, delta)
+        fullcmd = self._encode_cmd(cmd, key, str(delta), noreply)
         try:
-            server.send_cmd(cmd)
+            server.send_cmd(fullcmd)
+            if noreply:
+                return
             line = server.readline()
-            if line is None or line.strip() == 'NOT_FOUND':
+            if line is None or line.strip() == b'NOT_FOUND':
                 return None
             return int(line)
         except socket.error as msg:
@@ -579,7 +621,7 @@ class Client(threading.local):
             server.mark_dead(msg)
             return None
 
-    def add(self, key, val, time=0, min_compress_len=0):
+    def add(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Add new key with value.
 
         Like L{set}, but only stores in memcache if the key doesn't
@@ -588,9 +630,9 @@ class Client(threading.local):
         @return: Nonzero on success.
         @rtype: int
         '''
-        return self._set("add", key, val, time, min_compress_len)
+        return self._set("add", key, val, time, min_compress_len, noreply)
 
-    def append(self, key, val, time=0, min_compress_len=0):
+    def append(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Append the value to the end of the existing key's value.
 
         Only stores in memcache if key already exists.
@@ -599,9 +641,9 @@ class Client(threading.local):
         @return: Nonzero on success.
         @rtype: int
         '''
-        return self._set("append", key, val, time, min_compress_len)
+        return self._set("append", key, val, time, min_compress_len, noreply)
 
-    def prepend(self, key, val, time=0, min_compress_len=0):
+    def prepend(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Prepend the value to the beginning of the existing key's value.
 
         Only stores in memcache if key already exists.
@@ -610,9 +652,9 @@ class Client(threading.local):
         @return: Nonzero on success.
         @rtype: int
         '''
-        return self._set("prepend", key, val, time, min_compress_len)
+        return self._set("prepend", key, val, time, min_compress_len, noreply)
 
-    def replace(self, key, val, time=0, min_compress_len=0):
+    def replace(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Replace existing key with value.
 
         Like L{set}, but only stores in memcache if the key already exists.
@@ -621,9 +663,9 @@ class Client(threading.local):
         @return: Nonzero on success.
         @rtype: int
         '''
-        return self._set("replace", key, val, time, min_compress_len)
+        return self._set("replace", key, val, time, min_compress_len, noreply)
 
-    def set(self, key, val, time=0, min_compress_len=0):
+    def set(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Unconditionally sets a key to a given value in the memcache.
 
         The C{key} can optionally be an tuple, with the first element
@@ -643,7 +685,7 @@ class Client(threading.local):
         default to 0 == cache forever.
 
         @param min_compress_len: The threshold length to kick in
-        auto-compression of the value using the zlib.compress()
+        auto-compression of the value using the compressor
         routine. If the value being cached is a string, then the
         length of the string is measured, else if the value is an
         object, then the length of the pickle result is measured. If
@@ -652,10 +694,12 @@ class Client(threading.local):
         compatability, this parameter defaults to 0, indicating don't
         ever try to compress.
 
+        @param noreply: optional parameter instructs the server to not
+        send the reply.
         '''
-        return self._set("set", key, val, time, min_compress_len)
+        return self._set("set", key, val, time, min_compress_len, noreply)
 
-    def cas(self, key, val, time=0, min_compress_len=0):
+    def cas(self, key, val, time=0, min_compress_len=0, noreply=False):
         '''Check and set (CAS)
 
         Sets a key to a given value in the memcache if it hasn't been
@@ -678,7 +722,7 @@ class Client(threading.local):
         default to 0 == cache forever.
 
         @param min_compress_len: The threshold length to kick in
-        auto-compression of the value using the zlib.compress()
+        auto-compression of the value using the compressor
         routine. If the value being cached is a string, then the
         length of the string is measured, else if the value is an
         object, then the length of the pickle result is measured. If
@@ -686,14 +730,18 @@ class Client(threading.local):
         than the input, then it is discarded. For backwards
         compatability, this parameter defaults to 0, indicating don't
         ever try to compress.
+
+        @param noreply: optional parameter instructs the server to not
+        send the reply.
         '''
-        return self._set("cas", key, val, time, min_compress_len)
+        return self._set("cas", key, val, time, min_compress_len, noreply)
 
     def _map_and_prefix_keys(self, key_iterable, key_prefix):
         """Compute the mapping of server (_Host instance) -> list of keys to
         stuff onto that server, as well as the mapping of prefixed key
         -> original key.
         """
+        key_prefix = self._encode_key(key_prefix)
         # Check it just once ...
         key_extra_len = len(key_prefix)
         if key_prefix and self.do_check_key:
@@ -709,20 +757,39 @@ class Client(threading.local):
                 # Tuple of hashvalue, key ala _get_server(). Caller is
                 # essentially telling us what server to stuff this on.
                 # Ensure call to _get_server gets a Tuple as well.
-                str_orig_key = str(orig_key[1])
+                serverhash, key = orig_key
+
+                key = self._encode_key(key)
+                if not isinstance(key, six.binary_type):
+                    # set_multi supports int / long keys.
+                    key = str(key)
+                    if six.PY3:
+                        key = key.encode('utf8')
+                bytes_orig_key = key
 
                 # Gotta pre-mangle key before hashing to a
                 # server. Returns the mangled key.
                 server, key = self._get_server(
-                    (orig_key[0], key_prefix + str_orig_key))
+                    (serverhash, key_prefix + key))
+
+                orig_key = orig_key[1]
             else:
-                # set_multi supports int / long keys.
-                str_orig_key = str(orig_key)
-                server, key = self._get_server(key_prefix + str_orig_key)
+                key = self._encode_key(orig_key)
+                if not isinstance(key, six.binary_type):
+                    # set_multi supports int / long keys.
+                    key = str(key)
+                    if six.PY3:
+                        key = key.encode('utf8')
+                bytes_orig_key = key
+                server, key = self._get_server(key_prefix + key)
+
+            #  alert when passed in key is None
+            if orig_key is None:
+                self.check_key(orig_key, key_extra_len=key_extra_len)
 
             # Now check to make sure key length is proper ...
             if self.do_check_key:
-                self.check_key(str_orig_key, key_extra_len=key_extra_len)
+                self.check_key(bytes_orig_key, key_extra_len=key_extra_len)
 
             if not server:
                 continue
@@ -734,7 +801,8 @@ class Client(threading.local):
 
         return (server_keys, prefixed_to_orig_key)
 
-    def set_multi(self, mapping, time=0, key_prefix='', min_compress_len=0):
+    def set_multi(self, mapping, time=0, key_prefix='', min_compress_len=0,
+                  noreply=False):
         '''Sets multiple keys in the memcache doing just one query.
 
         >>> notset_keys = mc.set_multi({'key1' : 'val1', 'key2' : 'val2'})
@@ -777,7 +845,7 @@ class Client(threading.local):
             prefix not applied.
 
         @param min_compress_len: The threshold length to kick in
-            auto-compression of the value using the zlib.compress()
+            auto-compression of the value using the compressor
             routine. If the value being cached is a string, then the
             length of the string is measured, else if the value is an
             object, then the length of the pickle result is
@@ -786,6 +854,9 @@ class Client(threading.local):
             backwards compatability, this parameter defaults to 0,
             indicating don't ever try to compress.
 
+        @param noreply: optional parameter instructs the server to not
+            send the reply.
+
         @return: List of keys which failed to be stored [ memcache out
            of memory, etc. ].
 
@@ -809,21 +880,25 @@ class Client(threading.local):
                         mapping[prefixed_to_orig_key[key]],
                         min_compress_len)
                     if store_info:
-                        msg = "set %s %d %d %d\r\n%s\r\n"
-                        write(msg % (key,
-                                     store_info[0],
-                                     time,
-                                     store_info[1],
-                                     store_info[2]))
+                        flags, len_val, val = store_info
+                        headers = "%d %d %d" % (flags, time, len_val)
+                        fullcmd = self._encode_cmd('set', key, headers,
+                                                   noreply,
+                                                   b'\r\n', val, b'\r\n')
+                        write(fullcmd)
                     else:
                         notstored.append(prefixed_to_orig_key[key])
-                server.send_cmds(''.join(bigcmd))
+                server.send_cmds(b''.join(bigcmd))
             except socket.error as msg:
                 if isinstance(msg, tuple):
                     msg = msg[1]
                 server.mark_dead(msg)
                 dead_servers.append(server)
 
+        # if noreply, just return early
+        if noreply:
+            return notstored
+
         # if any servers died on the way, don't expect them to respond.
         for server in dead_servers:
             del server_keys[server]
@@ -853,16 +928,22 @@ class Client(threading.local):
         the new value itself.
         """
         flags = 0
-        if isinstance(val, str):
+        if isinstance(val, six.binary_type):
             pass
+        elif isinstance(val, six.text_type):
+            val = val.encode('utf-8')
         elif isinstance(val, int):
             flags |= Client._FLAG_INTEGER
-            val = "%d" % val
+            val = '%d' % val
+            if six.PY3:
+                val = val.encode('ascii')
             # force no attempt to compress this silly string.
             min_compress_len = 0
-        elif isinstance(val, long):
+        elif six.PY2 and isinstance(val, long):
             flags |= Client._FLAG_LONG
-            val = "%d" % val
+            val = str(val)
+            if six.PY3:
+                val = val.encode('ascii')
             # force no attempt to compress this silly string.
             min_compress_len = 0
         else:
@@ -878,11 +959,10 @@ class Client(threading.local):
             val = file.getvalue()
 
         lv = len(val)
-        # We should try to compress if min_compress_len > 0 and we
-        # could import zlib and this string is longer than our min
-        # threshold.
+        # We should try to compress if min_compress_len > 0
+        # and this string is longer than our min threshold.
         if min_compress_len and lv > min_compress_len:
-            comp_val = zlib.compress(val)
+            comp_val = self.compressor(val)
             # Only retain the result if the compression result is smaller
             # than the original.
             if len(comp_val) < lv:
@@ -896,7 +976,8 @@ class Client(threading.local):
 
         return (flags, len(val), val)
 
-    def _set(self, cmd, key, val, time, min_compress_len=0):
+    def _set(self, cmd, key, val, time, min_compress_len=0, noreply=False):
+        key = self._encode_key(key)
         if self.do_check_key:
             self.check_key(key)
         server, key = self._get_server(key)
@@ -906,26 +987,29 @@ class Client(threading.local):
         def _unsafe_set():
             self._statlog(cmd)
 
+            if cmd == 'cas' and key not in self.cas_ids:
+                return self._set('set', key, val, time, min_compress_len,
+                                 noreply)
+
             store_info = self._val_to_store_info(val, min_compress_len)
             if not store_info:
                 return(0)
+            flags, len_val, encoded_val = store_info
 
             if cmd == 'cas':
-                if key not in self.cas_ids:
-                    return self._set('set', key, val, time, min_compress_len)
-                fullcmd = "%s %s %d %d %d %d\r\n%s" % (
-                    cmd, key, store_info[0], time, store_info[1],
-                    self.cas_ids[key], store_info[2])
+                headers = ("%d %d %d %d"
+                           % (flags, time, len_val, self.cas_ids[key]))
             else:
-                fullcmd = "%s %s %d %d %d\r\n%s" % (
-                    cmd, key, store_info[0],
-                    time, store_info[1], store_info[2]
-                )
+                headers = "%d %d %d" % (flags, time, len_val)
+            fullcmd = self._encode_cmd(cmd, key, headers, noreply,
+                                       b'\r\n', encoded_val)
 
             try:
                 server.send_cmd(fullcmd)
-                return(server.expect("STORED", raise_exception=True)
-                       == "STORED")
+                if noreply:
+                    return True
+                return(server.expect(b"STORED", raise_exception=True)
+                       == b"STORED")
             except socket.error as msg:
                 if isinstance(msg, tuple):
                     msg = msg[1]
@@ -944,6 +1028,7 @@ class Client(threading.local):
             return 0
 
     def _get(self, cmd, key):
+        key = self._encode_key(key)
         if self.do_check_key:
             self.check_key(key)
         server, key = self._get_server(key)
@@ -954,7 +1039,9 @@ class Client(threading.local):
             self._statlog(cmd)
 
             try:
-                server.send_cmd("%s %s" % (cmd, key))
+                cmd_bytes = cmd.encode() if six.PY3 else cmd
+                fullcmd = b''.join((cmd_bytes, b' ', key))
+                server.send_cmd(fullcmd)
                 rkey = flags = rlen = cas_id = None
 
                 if cmd == 'gets':
@@ -973,7 +1060,7 @@ class Client(threading.local):
                 try:
                     value = self._recv_value(server, flags, rlen)
                 finally:
-                    server.expect("END", raise_exception=True)
+                    server.expect(b"END", raise_exception=True)
             except (_Error, socket.error) as msg:
                 if isinstance(msg, tuple):
                     msg = msg[1]
@@ -1072,7 +1159,8 @@ class Client(threading.local):
         dead_servers = []
         for server in six.iterkeys(server_keys):
             try:
-                server.send_cmd("get %s" % " ".join(server_keys[server]))
+                fullcmd = b"get " + b" ".join(server_keys[server])
+                server.send_cmd(fullcmd)
             except socket.error as msg:
                 if isinstance(msg, tuple):
                     msg = msg[1]
@@ -1087,7 +1175,7 @@ class Client(threading.local):
         for server in six.iterkeys(server_keys):
             try:
                 line = server.readline()
-                while line and line != 'END':
+                while line and line != b'END':
                     rkey, flags, rlen = self._expectvalue(server, line)
                     #  Bo Yang reports that this can sometimes be None
                     if rkey is not None:
@@ -1105,7 +1193,7 @@ class Client(threading.local):
         if not line:
             line = server.readline(raise_exception)
 
-        if line and line[:5] == 'VALUE':
+        if line and line[:5] == b'VALUE':
             resp, rkey, flags, len, cas_id = line.split()
             return (rkey, int(flags), int(len), int(cas_id))
         else:
@@ -1115,7 +1203,7 @@ class Client(threading.local):
         if not line:
             line = server.readline(raise_exception)
 
-        if line and line[:5] == 'VALUE':
+        if line and line[:5] == b'VALUE':
             resp, rkey, flags, len = line.split()
             flags = int(flags)
             rlen = int(len)
@@ -1134,15 +1222,22 @@ class Client(threading.local):
             buf = buf[:-2]  # strip \r\n
 
         if flags & Client._FLAG_COMPRESSED:
-            buf = zlib.decompress(buf)
+            buf = self.decompressor(buf)
+            flags &= ~Client._FLAG_COMPRESSED
 
-        if flags == 0 or flags == Client._FLAG_COMPRESSED:
-            # Either a bare string or a compressed string now decompressed...
-            val = buf
+        if flags == 0:
+            # Bare string
+            if six.PY3:
+                val = buf.decode('utf8')
+            else:
+                val = buf
         elif flags & Client._FLAG_INTEGER:
             val = int(buf)
         elif flags & Client._FLAG_LONG:
-            val = long(buf)
+            if six.PY3:
+                val = int(buf)
+            else:
+                val = long(buf)
         elif flags & Client._FLAG_PICKLE:
             try:
                 file = BytesIO(buf)
@@ -1173,27 +1268,26 @@ class Client(threading.local):
         """
         if isinstance(key, tuple):
             key = key[1]
-        if not key:
+        if key is None:
             raise Client.MemcachedKeyNoneError("Key is None")
+        if key is '':
+            if key_extra_len is 0:
+                raise Client.MemcachedKeyNoneError("Key is empty")
 
-        # Make sure we're not a specific unicode type, if we're old enough that
-        # it's a separate type.
-        if _has_unicode is True and isinstance(key, unicode):
-            raise Client.MemcachedStringEncodingError(
-                "Keys must be str()'s, not unicode.  Convert your unicode "
-                "strings using mystring.encode(charset)!")
-        if not isinstance(key, str):
-            raise Client.MemcachedKeyTypeError("Key must be str()'s")
+            #  key is empty but there is some other component to key
+            return
 
-        if isinstance(key, _str_cls):
-            if (self.server_max_key_length != 0 and
-                    len(key) + key_extra_len > self.server_max_key_length):
-                raise Client.MemcachedKeyLengthError(
-                    "Key length is > %s" % self.server_max_key_length
-                )
-            if not valid_key_chars_re.match(key):
-                raise Client.MemcachedKeyCharacterError(
-                    "Control/space characters not allowed (key=%r)" % key)
+        if not isinstance(key, six.binary_type):
+            raise Client.MemcachedKeyTypeError("Key must be a binary string")
+
+        if (self.server_max_key_length != 0 and
+                len(key) + key_extra_len > self.server_max_key_length):
+            raise Client.MemcachedKeyLengthError(
+                "Key length is > %s" % self.server_max_key_length
+            )
+        if not valid_key_chars_re.match(key):
+            raise Client.MemcachedKeyCharacterError(
+                "Control/space characters not allowed (key=%r)" % key)
 
 
 class _Host(object):
@@ -1241,7 +1335,7 @@ class _Host(object):
         self.socket = None
         self.flush_on_next_connect = 0
 
-        self.buffer = ''
+        self.buffer = b''
 
     def debuglog(self, str):
         if self.debug:
@@ -1284,7 +1378,7 @@ class _Host(object):
             self.mark_dead("connect: %s" % msg)
             return None
         self.socket = s
-        self.buffer = ''
+        self.buffer = b''
         if self.flush_on_next_connect:
             self.flush()
             self.flush_on_next_connect = 0
@@ -1296,10 +1390,14 @@ class _Host(object):
             self.socket = None
 
     def send_cmd(self, cmd):
-        self.socket.sendall(cmd + '\r\n')
+        if isinstance(cmd, six.text_type):
+            cmd = cmd.encode('utf8')
+        self.socket.sendall(cmd + b'\r\n')
 
     def send_cmds(self, cmds):
         """cmds already has trailing \r\n's applied."""
+        if isinstance(cmds, six.text_type):
+            cmds = cmds.encode('utf8')
         self.socket.sendall(cmds)
 
     def readline(self, raise_exception=False):
@@ -1312,10 +1410,10 @@ class _Host(object):
         if self.socket:
             recv = self.socket.recv
         else:
-            recv = lambda bufsize: ''
+            recv = lambda bufsize: b''
 
         while True:
-            index = buf.find('\r\n')
+            index = buf.find(b'\r\n')
             if index >= 0:
                 break
             data = recv(4096)
@@ -1333,8 +1431,11 @@ class _Host(object):
 
     def expect(self, text, raise_exception=False):
         line = self.readline(raise_exception)
-        if line != text:
-            self.debuglog("while expecting '%s', got unexpected response '%s'"
+        if self.debug and line != text:
+            if six.PY3:
+                text = text.decode('utf8')
+                line = line.decode('utf8', 'replace')
+            self.debuglog("while expecting %r, got unexpected response %r"
                           % (text, line))
         return line
 
@@ -1352,7 +1453,7 @@ class _Host(object):
 
     def flush(self):
         self.send_cmd('flush_all')
-        self.expect('OK')
+        self.expect(b'OK')
 
     def __str__(self):
         d = ''
@@ -1393,11 +1494,11 @@ if __name__ == "__main__":
                 return "%s (%s)" % (val, type(val))
             return "%s" % val
 
-        def test_setget(key, val):
+        def test_setget(key, val, noreply=False):
             global failures
-            print("Testing set/get {'%s': %s} ..."
-                  % (to_s(key), to_s(val)), end=" ")
-            mc.set(key, val)
+            print("Testing set/get (noreply=%s) {'%s': %s} ..."
+                  % (noreply, to_s(key), to_s(val)), end=" ")
+            mc.set(key, val, noreply=noreply)
             newval = mc.get(key)
             if newval == val:
                 print("OK")
@@ -1421,8 +1522,14 @@ if __name__ == "__main__":
                 return 0
 
         test_setget("a_string", "some random string")
+        test_setget("a_string_2", "some random string", noreply=True)
         test_setget("an_integer", 42)
-        if test_setget("long", long(1 << 30)):
+        test_setget("an_integer_2", 42, noreply=True)
+        if six.PY3:
+            ok = test_setget("long", 1 << 30)
+        else:
+            ok = test_setget("long", long(1 << 30))
+        if ok:
             print("Testing delete ...", end=" ")
             if mc.delete("long"):
                 print("OK")
@@ -1436,7 +1543,8 @@ if __name__ == "__main__":
                 print("FAIL")
                 failures += 1
         print("Testing get_multi ...",)
-        print(mc.get_multi(["a_string", "an_integer"]))
+        print(mc.get_multi(["a_string", "an_integer", "a_string_2",
+                            "an_integer_2"]))
 
         #  removed from the protocol
         # if test_setget("timed_delete", 'foo'):
@@ -1458,6 +1566,7 @@ if __name__ == "__main__":
 
         f = FooStruct()
         test_setget("foostruct", f)
+        test_setget("foostruct_2", f, noreply=True)
 
         print("Testing incr ...", end=" ")
         x = mc.incr("an_integer", 1)
@@ -1467,6 +1576,15 @@ if __name__ == "__main__":
             print("FAIL")
             failures += 1
 
+        print("Testing incr (noreply=True) ...", end=" ")
+        mc.incr("an_integer_2", 1, noreply=True)
+        x = mc.get("an_integer_2")
+        if x == 43:
+            print("OK")
+        else:
+            print("FAIL")
+            failures += 1
+
         print("Testing decr ...", end=" ")
         x = mc.decr("an_integer", 1)
         if x == 42:
@@ -1476,6 +1594,16 @@ if __name__ == "__main__":
             failures += 1
         sys.stdout.flush()
 
+        print("Testing decr (noreply=True) ...", end=" ")
+        mc.decr("an_integer_2", 1, noreply=True)
+        x = mc.get("an_integer_2")
+        if x == 42:
+            print("OK")
+        else:
+            print("FAIL")
+            failures += 1
+        sys.stdout.flush()
+
         # sanity tests
         print("Testing sending spaces...", end=" ")
         sys.stdout.flush()
@@ -1499,6 +1627,7 @@ if __name__ == "__main__":
         print("Testing using insanely long key...", end=" ")
         try:
             x = mc.set('a'*SERVER_MAX_KEY_LENGTH, 1)
+            x = mc.set('a'*SERVER_MAX_KEY_LENGTH, 1, noreply=True)
         except Client.MemcachedKeyLengthError as msg:
             print("FAIL")
             failures += 1
